package intermediate.symtab;

import java.util.ArrayList;

import intermediate.type.*;

/**
 * <h1>SymtabEntryImpl</h1>
 *
 * <p>An implementation of a symbol table entry.</p>
 *
 * <p>Copyright (c) 2020 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
public class SymtabEntry
{
    private String name;                     // entry name
    private Kind kind;                       // what kind of identifier
    private Symtab symtab;                   // parent symbol table
    private Typespec typespec;               // type specification
    private ArrayList<Integer> lineNumbers;  // source line numbers
    private EntryInfo info;                  // entry information
    
    /**
     * What kind of identifier.
     */
    public enum Kind
    {
        CONSTANT, ENUMERATION_CONSTANT, TYPE, VARIABLE, RECORD_FIELD,
        VALUE_PARAMETER, REFERENCE_PARAMETER, PROGRAM_PARAMETER,
        PROGRAM, PROCEDURE, FUNCTION,
        UNDEFINED;
        
        public String toString() { return super.toString().toLowerCase(); }
    }
    
    /**
     * Which routine.
     */
    public enum Routine
    {
        DECLARED, FORWARD,
        READ, READLN, WRITE, WRITELN,
        ABS, ARCTAN, CHR, COS, EOF, EOLN, EXP, LN, ODD, ORD,
        PRED, ROUND, SIN, SQR, SQRT, SUCC, TRUNC,
    }

    /**
     * Entry information interface.
     */
    private interface EntryInfo {}
    
    /**
     * Value information.
     */
    private class ValueInfo implements EntryInfo
    {
        private Object value;
    }
    
    /**
     * Constructor.
     * @param name the name of the entry.
     * @param kind the kind of entry.
     * @param symtab the symbol table that contains this entry.
     */
    public SymtabEntry(String name, Kind kind, Symtab symtab)
    {
        this.name = name;
        this.kind = kind;
        this.symtab = symtab;
        this.lineNumbers = new ArrayList<Integer>();

        // Initialize the appropriate entry information.
        switch (kind)
        {
            case CONSTANT:
            case ENUMERATION_CONSTANT:
            case VARIABLE:
            case RECORD_FIELD:
            case VALUE_PARAMETER:
                info = new ValueInfo();
                break;
            
            default: break;
        }
    }

    /**
     * Get the name of the entry.
     * @return the name.
     */
    public String getName() { return name; }
    
    /**
     * Get the kind of entry.
     * @return the kind.
     */
    public Kind getKind() { return kind; }
    
    /**
     * Set the kind of entry.
     * @param kind the kind to set.
     */
    public void setKind(Kind kind) { this.kind = kind; }

    /**
     * Get the symbol table that contains this entry.
     * @return the symbol table.
     */
    public Symtab getSymtab() { return symtab; }

    /**
     * Get the type specification of the entry.
     * @return the type specification.
     */
    public Typespec getType() { return typespec; }

    /**
     * Set the type specification.
     * @param typespec the type specification to set.
     */
    public void setType(Typespec typespec) { this.typespec = typespec; }

    /**
     * Get the arraylist of source line numbers for the entry.
     * @return the arraylist.
     */
    public ArrayList<Integer> getLineNumbers() { return lineNumbers; }

    /**
     * Append a source line number to the entry.
     * @param lineNumber the line number to append.
     */
    public void appendLineNumber(int lineNumber)
    {
        lineNumbers.add(lineNumber);
    }
    
    /**
     * Get the data value stored with this entry.
     * @return the data value.
     */
    public Object getValue() { return ((ValueInfo) info).value; }

    /**
     * Set the data value into this entry.
     * @parm value the value to set.
     */
    public void setValue(Object value) {  ((ValueInfo) info).value = value; }
}
